#!/usr/bin/perl -Tw
#=============================================================================
#
#        Copyright (C) 1999-2003, Matti Tukiainen, http://ktmatu.com/mt/
#
# Program name  : Blue dot
# Version       : 1.20
# Purpose       : Blue dot is a Perl CGI tracking and web site activity
#                   measurement script which generates Apache combined
#                   style access log files. These log files can be analyzed
#                   with most standard log analysis tools to track site's
#                   popularity, referrers, hosts, etc.
#
# Platforms     : All Perl platforms should be OK. Tested: Perl 5.6.1
#                   for sun4-solaris, Apache/1.3.27.
# Date          : 2003-06-03
# Updates       : 1.00 First release. (1999-11-18)
#                 1.01 Corrected a few documentation typos. (2001-05-30)
#                 1.10 Domain part of the URL is stripped to (2001-10-30)
#                        increase compatibility.
#                 1.20 Blue dot can be used as a Server Side (2003-06-03)
#                        Include (SSI) script. Client's IP, instead of
#                        proxy's IP, is logged, reverse DNS can be
#                        enabled, cookie based tracking added. HEAD
#                        requests are logged correctly.
#
# References    : NCSA extended/combined log format (TransferLog)
#                   <URL:http://hoohoo.ncsa.uiuc.edu/docs/setup/httpd/
#                   TransferLog.html>
#                   <URL:http://hoohoo.ncsa.uiuc.edu/docs/setup/httpd/
#                   LogOptions.html>
#                 Configuring Apache to use NCSA extended/combined log format
#                   <URL:http://www.apache.org/docs/mod/mod_log_config.html>
#
# Author        : Matti Tukiainen, <URL:http://ktmatu.com/mt/>
#
# Availability  : New releases are available at:
#                   <URL:http://ktmatu.com/software/bluedot/>
#
# Mailing list  : http://groups.yahoo.com/group/bluedot_announce/
#                 Send a blank email message to
#                 bluedot_announce-subscribe@yahoogroups.com to
#                 subscribe.
#
# Installation  : See the accompanying install.html file
#
# Copying       : This program may be used under the terms of GNU General
#                   Public License <URL:http://www.gnu.org/copyleft/gpl.html>
#
#=============================================================================

use strict;
use Socket;
use CGI::Carp qw(fatalsToBrowser);# Can be commented out if not available #!!!

#-----------------------------------------------------------------------------
# Name of the directory where the log files are written.
#-----------------------------------------------------------------------------
my ($LOG_DIR) = "/my/home/directory/logs/";

#-----------------------------------------------------------------------------
# It's impossible to get all necessary data from ENV alone. These values are
# used instead.
#-----------------------------------------------------------------------------
my ($DEFAULT_FILE_SIZE)        = 0;       # 0 bytes
my ($DEFAULT_HTTP_STATUS_CODE) = "200";   # OK
my ($DEFAULT_TIME_ZONE)        = "+0200"; # EET
my ($DEFAULT_DST_TIME_ZONE)    = "+0300"; # EET DST
my ($ENABLE_REVERSE_DNS)       = 1;       # 1 = slower, ml30pc3.uta.fi instead
                                          #     of 153.1.57.103
my ($ENABLE_COOKIE_LOG)        = 0;       # 1 = Generate Combined Format
                                          #     With Visitor Cookie Log




#-----------------------------------------------------------------------------
# Procedure parse_form
#
# Purpose  : Parses name-value pairs that were generated by GET or POST
#              method.
#
# Input    : None, input is read from the stdin.
# Output   : An assosiated array of variable names and values.
# Error
# handling : None
#
#-----------------------------------------------------------------------------
sub parse_form {
    my ($input) = "";
    my (%form);
    my ($name) = undef;
    my ($value) = undef;
    
    if ($ENV {'REQUEST_METHOD'} eq 'POST') {
        read (STDIN, $input, $ENV {'CONTENT_LENGTH'}); #!!! ?

        foreach (split (m/\&/, $input)) {
            next unless m/(.*?)\=(.*)/;
            $name = $1;
            $value = $2;

            $name =~ tr/+/ /;
            $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack ("C", hex ($1))/eg;

            $value =~ tr/+/ /;
            $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack ("C", hex ($1))/eg;

            if ($form {$name}) {
                $form {$name} .= ',' . $value;
            }
            else {
                $form {$name} = $value;
            }
        };
    }
    elsif (
        ($ENV {'REQUEST_METHOD'} eq 'GET')
        or
        ($ENV {'REQUEST_METHOD'} eq 'HEAD')
    ) {
        if (defined ($ENV {'QUERY_STRING'})) {
            $input = $ENV {'QUERY_STRING'};
        };

        foreach (split ('&', $input)) {
            s/\+/ /g;
            ($name, $value) = split ('=', $_, 2);
            $name =~ s/%([a-fA-F0-9]{2})/chr (hex ($1))/ge;
            $value =~ s/%([a-fA-F0-9]{2})/chr (hex ($1))/ge;

            # concatenate multiple vars
            if (defined ($form {$name})) {
                $form {$name} .= "\0" 
            }
            else {
                $form {$name} = "";
            };

            $form {$name} .= $value;
        };
    };

    return %form;
};




#-----------------------------------------------------------------------------
# Function get_cookies
#
# Purpose  : Returns a hash of cookie key, cookie value pairs.
#
# Input    : -
#
# Output   : Hash.
# Error
# handling : -
#
#-----------------------------------------------------------------------------
sub get_cookies {
    my ($buffer) = $ENV {'HTTP_COOKIE'} || "";
    my (@pairs)  = split(/; /, $buffer);
    my ($raw_cookie_str)   = split(/; /, $buffer);
    my ($cookie_name_str)  = "";
    my ($cookie_value_str) = "";
    my (%cookie_hash);

    foreach $raw_cookie_str (@pairs) {
        ($cookie_name_str, $cookie_value_str) =
            $raw_cookie_str =~ /(.*?)=(.*)/;

        $cookie_hash {$cookie_name_str} = $cookie_value_str;
    };

    return %cookie_hash;
};




#-----------------------------------------------------------------------------
# Function get_public_ip
#
# Purpose  : Gets a list of IP addresses and returns the first public
#              address from the list.
#
# Input    : @ip_list, List, list of IP address strings
#
# Output   : String, public IP.
# Error
# handling : Returns an empty string if there are no public IPs on the list.
#
#-----------------------------------------------------------------------------
sub get_public_ip {
    my (@ip_list) = @_;

    my ($ip) = "";
    my ($public_ip) = "";
    my ($num) = "";;

    foreach $ip (@ip_list) {
        $public_ip = $ip;

        # Private IPs
        if ($ip =~ /^(?:127|239|10|192\.168)\./) {
            $public_ip = "";
        }
        # Private IPs
        elsif ($ip =~ /^172\.(\d+)\./) {
            $num = $1;
            if (($num >= 16) && ($num <= 31)) {
                $public_ip = "";
            };
        };

        # Proxies sometimes return "unknown" as client's IP
        if ($ip !~ /\./) {
            $public_ip = "";
        };

        last if ($public_ip);
    };

    return $public_ip;
};




#-----------------------------------------------------------------------------
# Function number_to_domain
#
# Purpose  : Converts numeric IP address into human readable form.
#
# Input    : $a, $b, $c, $d, List, list of numbers in IP address
#
# Output   : String, converted IP address.
# Error
# handling : Returns a string of four numbers separated by periods if the
#              the address cannot be converted.
#
#-----------------------------------------------------------------------------
sub number_to_domain {
    my ($a, $b, $c, $d) = @_;
    my ($name, $aliases, $type, $len, @addrs);
    my ($addr) = pack ("C4", $a, $b, $c, $d);

    ($name, $aliases, $type, $len, @addrs) = gethostbyaddr ($addr, AF_INET);

    return $name ? $name : join (".", ($a,  $b, $c, $d));
};




#-----------------------------------------------------------------------------
# Function create_log_entry
#
# Purpose  : Concatenates different pieces of data and returns one log
#              line entry in Apache combined format.
#
# Input    : $time_now, current time returned by "time" function.
#            $public_ip, String, host name or address (example.com or 127.0...)
#            $location, String, requested document.
#            $referrer, String, referrer.
#            $file_size_int, Integer, number of bytes in the document.
#            $status_str, String, returned status code.
#            $normal_tz_str, String, normal time zone like "+0200".
#            $dst_tz_str, String, daylight savings time zone, like "+0300".
#            $cookie_str, String, cookies.
#
# Output   : $log_str, String, one log line.
# Error
# handling : default values are used if the required data is not available.
#
#-----------------------------------------------------------------------------
sub create_log_entry {
    my (
        $time_now,
        $public_ip,
        $location,
        $referrer,
        $protocol,
        $file_size_int,
        $status_str,
        $normal_tz_str,
        $dst_tz_str,
        $cookie_str
    ) = @_;

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
        localtime ($time_now);
    my (%month_name) = (
        1 => "Jan", 2 => "Feb", 3 => "Mar", 4 => "Apr", 5 => "May", 6 => "Jun",
        7 => "Jul", 8 => "Aug", 9 => "Sep", 10=> "Oct",11 => "Nov",12 => "Dec"
    );
    my ($time_zone_str) = ($isdst ?
       ($dst_tz_str || "+0001") :
       ($normal_tz_str || "+0000")
    );

    if ($cookie_str) {
        $cookie_str = " " . '"' . $cookie_str . '"';
    };

    my ($log_str) =
        $public_ip .
        " " .
        "-" .
        " " .
        "-" .
        " " .
        "[" .
            substr ("0" . $mday, -2) .
            "/" .
            $month_name {$mon + 1} .
            "/" .
            (1900 + $year) .
            ":" .
            substr ("0" . $hour, -2) .
            ":" .
            substr ("0" . $min, -2) .
            ":" .
            substr ("0" . $sec, -2) .
            " " .
            ($time_zone_str || "+0000") .
        "]" .
        " " .
        '"' .
        ($ENV {"REQUEST_METHOD"} || "UNKNOWN") .
        " " .
        $location .
        " " .
        $protocol .
        '"' .
        " " .
        ($status_str || "200") .
        " " .
        ($file_size_int || "0") .
        " " .
        '"' .
        $referrer .
        '"' .
        " " .
        '"' .
        ($ENV {"HTTP_USER_AGENT"} || "-") .
        '"' .
        $cookie_str;

    return $log_str;
};




#-----------------------------------------------------------------------------
# Function exclusive_append_to
#
# Purpose  : Appends one line to the end of locked file.
#
# Input    : $filename, string, name of the file.
#            @_, list of strings, items to be written to the file.
#
# Output   : -
# Error
# handling : Dies if something goes wrong.
#
#-----------------------------------------------------------------------------
sub exclusive_append_to {
    my $filename = shift;

    local (*HANDLE);

    open (HANDLE, ">>" . $filename) or die ("cannot append ".$filename.":".$!);
    # 2 = LOCK_EX
    flock (HANDLE, 2) or die ("Cannot flock: " . $filename . ":" . $!);
    seek (HANDLE, 0, 2) or die ("Cannot seek: " . $filename . ":" . $!);
    (print HANDLE @_) or die ("Cannot print: " . $filename . ":" . $!);
    close (HANDLE) or die ("Cannot close: " . $filename . ":" . $!);
};




#-----------------------------------------------------------------------------
# Function print_picture
#
# Purpose  : Sends response to a browser. A tiny 1x1 gif picture is the
#              default, if an URL of a picture is provided Location header
#              send.
#
# Input    : $picture_location, string, URL of the picture.  An empty
#              string means that the default picture is sent.
#
# Output   : -
# Error
# handling : -
# Notes    : $ od -tx1 -w8 -v mypicture.gif
#            can be used to convert binary file into suitable format.
#
#-----------------------------------------------------------------------------
sub print_picture {
    my $picture_location = shift;
    my $ssi_used = shift;
    my $pid_cookie = shift;
    my $sid_cookie = shift;

    my (@dot_gif) = (
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x01, 0x00,
        0x01, 0x00, 0xb3, 0x00, 0x00, 0x00, 0x00, 0x63,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0x2c, 0x00, 0x00,
        0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x04,
        0x02, 0x10, 0x44, 0x00, 0x3b
    );

    if (($picture_location ne "") && ($picture_location =~ m|^http://|)) {
        print ("Location: " . $picture_location . "\n");

        if ($sid_cookie) {
            (print "Set-Cookie: " . "SID=" . $sid_cookie . "; " .
                "path=/;\n");
        };
        if ($pid_cookie) {
            (print "Set-Cookie: " . "PID=" . $pid_cookie . "; " .
                "path=/; expires=Thu, 22 Dec 2022 22:22:22 GMT\n");
        };

        print ("\n");
    }
    elsif ($ssi_used) {
        # SSI
        (print
            "Content-type: text/plain\n\n" .
            ""
        );
    }
    else {
        if ($sid_cookie) {
            print ('P3P: CP="NOI ADM DEV PSAi COM NAV OUR OTRo STP IND DEM"');
            print ("\n");
            (print "Set-Cookie: " . "SID=" . $sid_cookie . "; " .
                "path=/;\n");
        };
        if ($pid_cookie) {
            (print "Set-Cookie: " . "PID=" . $pid_cookie . "; " .
                "path=/; expires=Thu, 22 Dec 2022 22:22:22 GMT\n");
        };

        (print
            "Content-type: image/gif\n\n" .
            join ("", map (chr ($_), @dot_gif))
        );
    };
};




sub main {
    my (%form_vars)     = parse_form ();
    my ($ssi_used)      = $form_vars {"ssi"} ? 1 : 0;
    my ($time_now)      = time;
    my ($referrer)      = "";
    my ($location)      = "";
    my ($image)         = ($form_vars {"image"} || "");
    my ($log_id)        = ($form_vars {"id"}    || "bluedotlog");
    my ($protocol)      = ($ENV {"SERVER_PROTOCOL"} || "UNKNOWN");
    my (%cookies)       = get_cookies ();
    my ($pid_cookie)    = "";  # Persistent ID
    my ($sid_cookie)    = "";  # Session ID
    my ($log_filename)  = "";


    my ($ip_behind_proxy) = (
        $ENV {"HTTP_X_FORWARDED_FOR"} ||
        $ENV {"HTTP_CLIENT_IP"}       ||
        ""
    );

    my ($public_ip) = get_public_ip (
        split (/, */, $ip_behind_proxy),
        $ENV {"REMOTE_HOST"},
        $ENV {"REMOTE_ADDR"},
        "255.255.255.255"
    );

    if ($ENABLE_REVERSE_DNS) {
        if ($public_ip =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/o) {
            $public_ip = number_to_domain (split (/\./, $1));
        };
        if ($public_ip !~ /\./) {
            $public_ip = $ENV {"REMOTE_ADDR"};
        };
    };

    $log_id =~ /(.*)/; # untaint
    $log_id = $1;
    $log_id = lc ($log_id);
    $log_id =~ s/[^a-z]//g;

    if ($log_id eq "") {
        $log_id = "bluedotlog";
    };

    $log_filename = $LOG_DIR . "bluedotlog";
    if (-e ($LOG_DIR . $log_id)) {
        $log_filename = $LOG_DIR . $log_id;
    };

    if ($protocol =~ /^INCLUDED$/) {
        # Apache uses INCLUDED if SSI is used
        $ssi_used = 1;
        $protocol = "HTTP/1.1";
    };

    if ($ssi_used) {
        $referrer = ($ENV {"HTTP_REFERER"} || "-");
        $location = ($ENV {"REQUEST_URI"} || "UNKNOWN");
    }
    else {
        $referrer = ($form_vars {"referrer"} || "-");
        $location = (
            $form_vars {"location"} || $ENV {"HTTP_REFERER"} || "UNKNOWN"
        );
        $referrer =~ s/ /+/g;

	$location =~ s/ /+/g;
	($location) = $location =~ m|^.*?://.*?/(.*)|i;
	$location = "/" . ($location || "");
    };

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
        localtime ($time_now);
    my ($id_time) = (1900 + $year) . substr ("0" . ($mon + 1), -2) .
        substr ("0" . $mday, -2) . substr ("0" . $hour, -2) .
        substr ("0" . $min, -2) . substr ("0" . $sec, -2);

    my ($cookie_log) = "";

    if ($ENABLE_COOKIE_LOG) {
	my ($rnd_str) = chr (int (ord ("a") + rand (ord ("z") - ord ("a"))));
	$rnd_str .= chr (int (ord ("a") + rand (ord ("z") - ord ("a"))));
	$rnd_str .= chr (int (ord ("a") + rand (ord ("z") - ord ("a"))));

        $pid_cookie = $cookies {"PID"} || ($id_time . "-" . $rnd_str);
        $sid_cookie = $cookies {"SID"} || ($id_time . "-" . $rnd_str);
        $cookie_log = "PID=" . $pid_cookie . ";SID=" . $sid_cookie;
    };

    exclusive_append_to (
        $log_filename,
        create_log_entry (
            $time_now,
            $public_ip,
            $location,
            $referrer,
            $protocol,
            $DEFAULT_FILE_SIZE,
            $DEFAULT_HTTP_STATUS_CODE,
            $DEFAULT_TIME_ZONE,
            $DEFAULT_DST_TIME_ZONE,
            $cookie_log
        ),
        "\n"
    );

    print_picture ($image, $ssi_used, $pid_cookie, $sid_cookie);
};




main ();
